\ Copyright (c) 2002 by eForth Technology Inc.
\ All rights reserved 
\
\ GBAKERL.F   18MAY02 created by Wuja.
\
\ GBAeFORTH low-level definitions.
\
\ (Translated from eForth 1.0 by Bill Muench and C. H. Ting, 1990)
\ 
\ V1.00,   24OTC01WUJA,  Translate low-level defintions from eForth 1.0.
\ V1.10,   14DEC01WUJA,  Branch to AVNET's BK100PHTB.S V1.0 platform.
\ V1.20,   18MAY02WUJA,  Branch to Nitendo's GBA platform.
\ V2.01,   26SEP04CHT,   Port to weForth
\

CODE ?RX   ( -- c T | F )                   \ Return input character and true, or a false if no input.
  WP, $F800 #               	,MVN        \ WP := [PC] (0x04000100)
  [-P] XP, WP, $EB #            ,LDR        \ Get SIOCNT register
  S, XP, XP, $1 #               ,AND        \ get SIO_DATA_EMPTY 
if
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  [-P] TOS, WP, $FF #           ,LDR        \ get char from SIODATA8 register
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, 0 #                      ,MVN        \ TOS := TRUE
  $NEXT
then
  [-P!] TOS, SP, 4 #            ,STR        \ SP := SP - 4
  TOS, 0 #                      ,MOV        \ TOS := FALSE
  $NEXT
END-CODE  

CODE TX!   ( c -- )                         \ Send character c to the output device.
begin
  WP, $F800 #               	,MVN        \ WP := [PC] (0x04000100)
  [-P] XP, WP, $EB #            ,LDR        \ Get SIOCNT register
  S, XP, XP, $20 #              ,AND        \ get SIO_DATA_EMPTY 
until
  [-P] TOS, WP, $FF #           ,STR        \ Send char to SIODATA8 register
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  $NEXT
END-CODE

CODE !IO   ( -- )                           \ Initialize the serial I/O devices.
  WP, $BF0 #                	,MVN        \ GP1CON
  XP, $11 #                   	,MOV        \ tx+P1.0, rx=P1.1
  [-P] XP, WP, $B #             ,STR        \ 
  WP, $F800 #                	,MVN        \ COMSTA0
  XP, $80 #                   	,MOV        \ DLAB access COMDIV0/COMDIV1
  [-P] XP, WP, $F3 #            ,STR        \ 
\  XP, $18 #                   	,MOV        \ COMDIV0 ADuc7020 Rev. E
  XP, $16 #                   	,MOV        \ COMDIV0 ADuC7020 Rev. H
  [-P] XP, WP, $FF #            ,STR        \ 
  XP, 0 #                   	,MOV        \ COMDIV1
  [-P] XP, WP, $FB #            ,STR        \ 
  XP, $7 #                   	,MOV        \ 8 data, 2 stop bits
  [-P] XP, WP, $F3 #            ,STR        \ 
  $NEXT
END-CODE
                 
\ The ARM7EF kernel

CODE NOP  ( -- w )                        \ Push an inline literal.
  $NEXT
END-CODE

CODE doLIT  ( -- w )                        \ Push an inline literal.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  [] TOS, LP, 4 #               ,LDR        \ TOS := [LP], LP := LP+4
  $NEXT
END-CODE COMPILE-ONLY

CODE EXECUTE ( ca -- )                      \ Execute the word at ca.
  WP, TOS,                      ,MOV        \ get next CFA
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  WP                            ,BX         \ jump to next CFA

CODE next ( -- )                            \ Run time code for the single index loop.
  [P] WP, RP                     ,LDR        \ get current index
  S, WP, WP,                    ,MOV        \ WP := WP
ifz
  RP, RP, $4 #                  ,ADD        \ RP := RP+4
  LP, LP, $4 #                  ,ADD        \ LP := LP+4
  $NEXT
then
  WP, WP, 1 #                   ,SUB        \ WP := WP -1
  [P] WP, RP                     ,STR        \ [RP] := WP
  [P] LP, LP                     ,LDR        \ LP := [LP]
  $NEXT
END-CODE COMPILE-ONLY

CODE ?branch  ( f -- )                      \ Branch if flag is zero.
  S, WP, TOS,                   ,MOV        \ get f
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
ifz
  [P] LP, LP                     ,LDR        \ LP := [LP]
  $NEXT
then
  LP, LP, $4  #                 ,ADD        \ LP := LP+4
  $NEXT
  END-CODE COMPILE-ONLY

CODE branch  ( -- )                         \ Branch to an inline address.
  [P] LP, LP                     ,LDR        \ LP := [LP]
  $NEXT
END-CODE COMPILE-ONLY

CODE ! ( w a -- )                           \ Pop the data stack to memory.
  [] WP, SP, 4 #                ,LDR        \ get w, SP := SP+4
  [P] WP, TOS                    ,STR        \ a := w
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  $NEXT
END-CODE  

CODE @ ( a -- w )                           \ Push memory location to the data stack.
  [P] TOS, TOS                   ,LDR        \ TOS := [TOS]
  $NEXT
END-CODE

CODE W! ( w a -- )                          \ Pop the data stack to half word memory.
  [] WP, SP, 4 #                ,LDR        \ get w, SP := SP+4
  [] WP, TOS, 0 #               ,STRH        \ a := w
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  $NEXT
END-CODE

CODE W@ ( a -- w )                           \ Push half word to the data stack.
  [P] TOS, TOS, 0 #              ,LDRH        \ TOS := [TOS]
  $NEXT
END-CODE

CODE C! ( c b -- )                          \ Pop the data stack to byte memory.
  [] WP, SP, 4 #                ,LDR        \ get c, SP := SP+4
  [PB] WP, TOS                   ,STR        \ b := c
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  $NEXT
END-CODE

CODE C@ ( b -- c )                          \ Push byte memory location to the data stack.
  [PB] TOS, TOS                  ,LDR        \ TOS := byte [TOS]
  $NEXT
END-CODE

CODE RP@ ( -- a )                           \ Push the current RP to the data stack.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, RP                       ,MOV        \ TOS := RP
  $NEXT
END-CODE  

CODE RP! ( a -- )                           \ Set the return stack pointer.
  RP, TOS                       ,MOV        \ RP := TOS
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  $NEXT
END-CODE COMPILE-ONLY

CODE R> ( -- w )                            \ Pop the return stack to the data stack.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  [] TOS, RP, 4 #               ,LDR        \ pop TOS from RS
  $NEXT
END-CODE

CODE R@ ( -- w )                            \ Copy top of return stack to the data stack.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  [P] TOS, RP                    ,LDR        \ TOS := [RP]
  $NEXT
END-CODE  

CODE >R ( w -- )                            \ Push the data stack to the return stack.
  [-P!] TOS, RP, 4 #            ,STR        \ push TOS to RS
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  $NEXT  
END-CODE COMPILE-ONLY

CODE SP@ ( -- a )                           \ Push the current data stack pointer.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, SP                       ,MOV        \ TOS := SP
  $NEXT
END-CODE  

CODE SP! ( a -- )                           \ Set the data stack pointer.
  SP, TOS                       ,MOV        \ SP := TOS
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  $NEXT
END-CODE

CODE DROP ( w -- )                          \ Discard top stack item.
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  $NEXT
END-CODE

CODE DUP ( w -- w w )                       \ Duplicate the top stack item.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  $NEXT
END-CODE  

CODE SWAP ( w1 w2 -- w2 w1 )                \ Exchange top two stack items.
  [P] WP, SP                     ,LDR        \ WP := [SP]
  [P] TOS, SP                    ,STR        \ [SP] := TOS
  TOS, WP                       ,MOV        \ TOS := WP
  $NEXT
END-CODE

CODE OVER ( w1 w2 -- w1 w2 w1 )             \ Copy second stack item to top.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  [P] TOS, SP, 4 #              ,LDR        \ TOS := [SP+4] 
  $NEXT
END-CODE  

CODE 0< ( n -- t )                          \ Return true if n is negative.
  WP, 0 #                       ,MOV        \ WP := 0
  TOS, WP, TOS, $20 ASR#         ,ADD        \ TOS := 0 + (0|-1)
  $NEXT
END-CODE

CODE AND ( w1 w2 -- w )                     \ Bitwise AND.
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  TOS, TOS, WP                  ,AND        \ TOS := TOS and WP
  $NEXT
END-CODE  

CODE OR ( w w -- w )                        \ Bitwise inclusive OR.
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  TOS, TOS, WP                  ,ORR        \ TOS := TOS or WP
  $NEXT
END-CODE  

CODE XOR ( w w -- w )                       \ Bitwise exclusive OR.
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  TOS, TOS, WP                  ,EOR        \ TOS := TOS xor WP
  $NEXT
END-CODE  

CODE RSHIFT ( w1 w2 -- w )                     \ Bitwise AND.
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  TOS, WP, TOS, TOS, ASR             ,MOV        \ TOS := TOS and WP
  $NEXT
END-CODE  

CODE LSHIFT ( w1 w2 -- w )                     \ Bitwise AND.
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  TOS, WP, TOS, TOS, LSL             ,MOV        \ TOS := TOS and WP
  $NEXT
END-CODE  

CODE UM+ ( w1 w2 -- w cy )                  \ Add two numbers, return the sum and carry flag.
  [P] WP, SP                     ,LDR        \ WP = [SP]
  S, WP, WP, TOS                ,ADD        \ WP := WP ADD TOS
  TOS, 0 #                      ,MOV        \ TOS := 0
  TOS, TOS, 0 #                 ,ADC        \ TOS := TOS+Carry
  [P] WP, SP                     ,STR        \ [SP] := WP
  $NEXT 
END-CODE  

CODE + ( w1 w2 -- w )                     \ Bitwise AND.
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  TOS, TOS, WP                  ,ADD        \ TOS := TOS and WP
  $NEXT
END-CODE  

CODE - ( w1 w2 -- w )                     \ Bitwise AND.
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  TOS, TOS, WP                  ,RSB        \ TOS := TOS and WP
  $NEXT
END-CODE  

CODE * ( w1 w2 -- w )                     \ Bitwise AND.
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  TOS, TOS, WP                  ,MUL        \ TOS := TOS and WP
  $NEXT
END-CODE  

CODE UM* ( u1 u2 -- udl udh )               \ Unsigned multiply. Return double product.
  [P] WP, SP                     ,LDR        \ WP := u1
  XP, YP, TOS, WP              ,UMULL      \ TOS:XP := TOS*WP
  [P] XP, SP                     ,STR        \ [SP] := XP(udl)
  TOS, YP			,MOV
  $NEXT
END-CODE  

CODE M* ( w1 w2 -- lo hi ) 
  [P] WP, SP			,LDR    
  XP, YP, TOS, WP		,SMULL  
  [P] XP, SP			,STR    
  TOS, YP			,MOV
  $NEXT
END-CODE  

CODE 1+ ( w  -- w+1 )                     \ Bitwise AND.
  TOS, TOS, 1 #                  ,ADD        \ TOS := TOS and WP
  $NEXT
END-CODE  

CODE 1- ( w  -- w-1 )                     \ Bitwise AND.
  TOS, TOS, 1 #                  ,SUB        \ TOS := TOS and WP
  $NEXT
END-CODE  

CODE 2+ ( w  -- w+2 )                     \ Bitwise AND.
  TOS, TOS, 2 #                  ,ADD        \ TOS := TOS and WP
  $NEXT
END-CODE  

CODE 2- ( w  -- w-2 )                     \ Bitwise AND.
  TOS, TOS, 2 #                  ,SUB        \ TOS := TOS and WP
  $NEXT
END-CODE  

CODE CELL+ ( w  -- w+4 )                     \ Bitwise AND.
  TOS, TOS, 4 #                  ,ADD        \ TOS := TOS and WP
  $NEXT
END-CODE  

CODE CELL- ( w  -- w-4 )                     \ Bitwise AND.
  TOS, TOS, 4 #                  ,SUB        \ TOS := TOS and WP
  $NEXT
END-CODE  

CODE BL ( w -- w w )                       \ Duplicate the top stack item.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, $20 #                    ,MOV
  $NEXT
END-CODE  

CODE CELLS ( n -- t )                          \ Return true if n is negative.
   TOS, TOS 2 LSL#              ,MOV        \ WP := 0
  $NEXT
END-CODE

CODE CELL/ ( n -- t )                          \ Return true if n is negative.
  TOS, TOS, 2 ASR#          ,MOV       \ TOS := TOS*2
  $NEXT
END-CODE

CODE 2* ( w  -- w+2 )                     \ Bitwise AND.
   TOS, TOS, 1 LSL#              ,MOV        \ WP := 0
  $NEXT
END-CODE  

CODE 2/ ( n -- t )                          \ Return true if n is negative.
   TOS, TOS, 1 ASR#              ,MOV        \ WP := 0
  $NEXT
END-CODE

CODE 8LSHIFT ( n -- t )                          \ Return true if n is negative.
  TOS, TOS, 8 LSL#          ,MOV       \ TOS := 0 + TOS*256
  $NEXT
END-CODE

CODE 16LSHIFT ( n -- t )                          \ Return true if n is negative.
  TOS, TOS, 10 LSL#          ,MOV       \ TOS := 0 + TOS*256
  $NEXT
END-CODE

CODE 24LSHIFT ( n -- t )                          \ Return true if n is negative.
  TOS, TOS, 18 LSL#          ,MOV       \ TOS := 0 + TOS*256
  $NEXT
END-CODE

CODE 8RSHIFT ( n -- t )                          \ Return true if n is negative.
  TOS, TOS, 8 LSR#          ,MOV       \ TOS := TOS*$100
  $NEXT
END-CODE

CODE 16RSHIFT ( n -- t )                          \ Return true if n is negative.
  TOS, TOS, 10 LSR#          ,MOV       \ TOS := TOS*$10000
  $NEXT
END-CODE

CODE 24RSHIFT ( n -- t )                          \ Return true if n is negative.
  TOS, TOS, 18 LSR#          ,MOV       \ TOS := TOS*$1000000
  $NEXT
END-CODE

CODE ?DUP ( w -- w w | 0 )                  \ Duplicate the top stack item.
  S, WP, TOS,                   ,MOV        \ get f
  NE, [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  $NEXT
  END-CODE

CODE ROT ( w1 w2 w3 -- w2 w3 w1 )           \ Rotate top three stack items.
  [P] WP, SP                     ,LDR        \ WP := w2
  [P] TOS, SP                    ,STR        \ push w3
  [P] TOS, SP, 4 #              ,LDR        \ TOS := w1 
  [P] WP, SP, 4 #               ,STR        \ push w2
  $NEXT
END-CODE

CODE 2DROP ( w w -- )                       \ Discard top stack item.
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  $NEXT
END-CODE

CODE 2DUP ( w1 w2 -- w1 w2 w1 w2 )          \ Copy two stack item to top.
  [P] WP, SP                     ,LDR        \ WP := w1
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  [-P!] WP, SP, 4 #             ,STR        \ push TOS
  $NEXT
END-CODE  

CODE D+ ( d1 d2 -- d3 )
  [] WP, SP, 4 #                ,LDR        \ WP = d2low
  [] XP, SP, 4 #                ,LDR        \ XP = d1high
  [P] YP, SP	                ,LDR        \ YP = d1low
  S, WP, WP, YP                	,ADD        \ d1low+d2loq
  [P] WP, SP                     ,STR        \ store d3low
  TOS, TOS, XP                 	,ADC        \ d3high
  $NEXT 
END-CODE  

CODE NOT ( n -- t )                          \ Return 1's complement
  TOS, TOS 	            	,MVN         \ WP := 0
  $NEXT
END-CODE

CODE NEGATE ( n -- t )                      \ Return negative.
  TOS, TOS, 0 #	            	,RSB        \ WP := 0
  $NEXT
END-CODE

CODE ABS ( n -- t )                        \ Return absolute
  S, TOS, 80000000 #            ,TST        \ get n
  NE, TOS, TOS, 0 #		,RSB
  $NEXT
END-CODE

CODE = ( w w -- t ) 
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  S, WP, TOS               	,CMP        \ TOS := TOS xor WP
  EQ, TOS, 0 #			,MVN
  NE, TOS, 0 #			,MOV
  $NEXT
END-CODE  

CODE U< ( w w -- t ) 
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  S, WP, TOS     	        ,CMP        \ WP-TOS
  CC, TOS, 0 #			,MVN
  CS, TOS, 0 #			,MOV
  $NEXT
END-CODE  

CODE < ( w w -- t ) 
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  S, WP, TOS     	        ,CMP        \ WP-TOS
  LT, TOS, 0 #			,MVN
  GE, TOS, 0 #			,MOV
  $NEXT
END-CODE  

CODE MAX ( w w -- t ) 
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  S, WP, TOS     	        ,CMP        \ WP-TOS
  GT, TOS, WP			,MOV
  $NEXT
END-CODE  

CODE MIN ( w w -- t ) 
  [] WP, SP, 4 #                ,LDR        \ WP = [SP], SP := SP+4
  S, WP, TOS     	        ,CMP        \ WP-TOS
  LT, TOS, WP			,MOV
  $NEXT
END-CODE  

CODE +! ( w a -- )                           \ Pop the data stack to memory.
  [] WP, SP, 4 #                ,LDR        \ get w, SP := SP+4
  [P] XP, TOS                    ,LDR        \ a := w
  XP, XP, WP                 	,ADD        \ a := w
  [P] XP, TOS                    ,STR        \ a := w
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  $NEXT
END-CODE  

CODE 2! ( d a -- )                           \ Pop the data stack to memory.
  [] WP, SP, 4 #                ,LDR        \ get hi, SP := SP+4
  [] XP, SP, 4 #                ,LDR        \ get low, SP := SP+4
  [] WP, TOS, 4 #               ,STR        \ a := w
  [P] XP, TOS			,STR        \ a := w
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  $NEXT
END-CODE  

CODE 2@ ( a -- d )                           \ Push memory location to the data stack.
  [+P] WP, TOS, 4 #               ,LDR        \ TOS := [TOS]
  [-P!] WP, SP, 4 #            ,STR        \ push TOS
  [P] TOS, TOS                   ,LDR        \ TOS := [TOS]
  $NEXT
END-CODE

CODE COUNT ( b -- b+1 c )                          \ Push byte memory location to the data stack.
  [+B] WP, TOS, 1 #            ,LDR        \ TOS := byte [TOS]
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, WP                       ,MOV        \ TOS := RP
  $NEXT
END-CODE

CODE DNEGATE ( d -- -d )
  [P] WP, SP                     ,LDR        \ WP = [SP]
  S, WP, WP, 0 #	        ,RSB        \ WP := WP ADD TOS
  TOS, TOS, 0 #                 ,RSC        \ TOS := TOS+Carry
  [P] WP, SP                     ,STR        \ [SP] := WP
  $NEXT 
END-CODE  

